#!/usr/bin/python
#
# Copyright 2015 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import dateutil.parser
import os
import re
import string
import syslog
import time
import azsec.util.common as util

from defusedxml.ElementTree import parse

defaultFacility = syslog.LOG_AUTH
# ClamAV Update options
sigTypes = [ "main", "daily", "bytecode" ]
imsResponse = "304 Not Modified"
tmpDir = "/tmp"
connectTimeout = "30"
downloadAttempts = "3"
defaultConfigFile = "{0}/{1}".format(util.pkgConfigDir, "plugins/clamav/avupdate.xml")
signatureMaxAge = 24

# ClamAV Scanning Options
defaultScanDir="/"
excludeDir="/sys/"
scanName = 'avscan'
scanDir = 'avscandir'

argRootDir = "targetDir"
argConfigFile = "avConfigFile"

def ParseConfig(configFile):
    if not configFile:
        util.WriteConsole(syslog.LOG_ERR, "Config file path must be a valid string")
        return False
    if not os.path.isfile(configFile):
        util.WriteConsole(syslog.LOG_ERR, "No config file exists at the path {0}".format(configFile))
        return False
    # Parse the file
    try:
        util.WriteVerbose("Parsing config file " + configFile)
        elementTree = parse(configFile)
        root = elementTree.getroot()
        options = root.find('options')
    except:
        util.WriteConsole(syslog.LOG_ERR, "Unable to parse file as XML: " +  configFile)
        return False
    global urlPath, proxyServer, sigDir, sigOwner, sigPerms
    urlPath = options.get('urlPath')
    proxyServer = options.get('proxyServer')
    sigDir = options.get('sigDir')
    sigOwner = options.get('sigOwner')
    sigPerms = options.get('sigPerms')
    return True

def CheckSignatures():
    # Get Signature date
    err,msg = util.RunGetOutput("clamscan -V")
    if (err > 1):
        util.WriteConsole(syslog.LOG_ERR, "Error running clamscan -V to determine signature currency")
        return False
    fields = string.split(msg, "/")
    # if (fields.length < 3)
    sigTime = dateutil.parser.parse(fields[2])
    now = datetime.datetime.now()
    elapsedTime = now - sigTime
    elapsedHours = elapsedTime.total_seconds() / 3600
    util.WriteVerbose("System Time: {0}, Signature time: {1}, Signatures are {2} hours old".format(now, sigTime, elapsedHours))
    if (elapsedHours > signatureMaxAge):
        util.WriteLog(syslog.LOG_WARNING, "Antivirus signatures are out of date ({0} hours old)".format(elapsedHours))
        return False
    util.WriteLog(syslog.LOG_INFO, "Antivirus signatures are up to date ({0} hours old)".format(elapsedHours))
    return True
    
# If file exists, return true and the rfc 1123 formatted filetime
# Otherwise return false
def GetFileTime(file):
    if not file:
        return False, "Must supply a valid filename"
    if not os.path.isfile(file):
        return False, "No file exists at path " + file
    # Get file modification time and convert to GMT
    fileTime = os.path.getmtime(file)
    gmTime = time.gmtime(fileTime)
    # Format per rfc 1123
    formattedTime = time.strftime("%a, %d %b %Y %H:%M:%S GMT", gmTime)
    return True, formattedTime

# Attempt to download specified type of signature
# Return true if downloaded successfully or if 304 received
# Return false if any failure during download
def UpdateSignature(sigType):
    print "Fetching {0} signatures".format(sigType)
    global urlPath, proxyServer, sigDir, sigOwner, sigPerms
    fileName = "{0}.cvd".format(sigType)
    tmpPath = "{0}/{1}".format(tmpDir, fileName)
    filePath = "{0}/{1}".format(sigDir, fileName)
    success, formattedTime = GetFileTime(filePath)
    params = "--no-check-certificate"
    # If we could get a file time, include an IMS header
    if success:
        params += " --header 'If-Modified-Since: {0}'".format(formattedTime)
    # If proxy is configured, use it
    if proxyServer:
        params += " -e http_proxy='{0}'".format(proxyServer)
    else:
        proxyOption = ""
    params += " --connect-timeout={0} --tries={1}".format(connectTimeout, downloadAttempts)
    params += " -O '{0}'".format(tmpPath)
    url = "{0}/{1}".format(urlPath, fileName)
    command = "wget {0} {1} 2>&1".format(params, url)
    err,msg = util.RunGetOutput(command, False)
    # 304's result in a non-zero error code, so return true immediately if 304 is found in response
    match = re.search(imsResponse, msg)
    if (match != None):
        return True, "Local file ({0}) is already up to date".format(filePath)
    if (err != 0):
        return False, "Unable to retrieve {0}".format(url)
    # Successfully retrieved file, now move it to te destination folder
    command = "chown {1} {0}; chmod {2} {0}; mv {0} {3}".format(tmpPath, sigOwner, sigPerms, sigDir)
    err, msg = util.RunGetOutput(command)
    if (err != 0):
        return False, "Retrieved signature from {0} but unable to move it to {1}".format(url, sigDir)
    return True, "Successfully retrieved update from {0} to {1}".format(url, sigDir)
            
def UpdateSignatures(configFile = defaultConfigFile):
    util.WriteConsole(syslog.LOG_INFO, "Begin check for ClamAV signatures")
    # Parse Config File
    if not ParseConfig(configFile):
        return 1
    errors = 0
    summaryLevel = syslog.LOG_INFO
    for sigType in sigTypes:
        success, result = UpdateSignature(sigType)
        if not success:
            errors += 1
            summaryLevel = syslog.LOG_WARNING
            level = syslog.LOG_WARNING
        else:
            level = syslog.LOG_INFO
        util.WriteConsole(level, result)
    util.WriteConsole(summaryLevel, "Finished execution with {0} errors".format(errors))
    if not CheckSignatures():
        errors += 1
    return errors
    
def RunAVScan(target = defaultScanDir):
    commands = {}
    commands[util.ubuntu] = "nice -n 5 clamscan -r -i --exclude-dir={0} {1}".format(excludeDir, target)
    if not commands.has_key(util.distro):
        util.WriteLog(syslog.LOG_ERR, "{0} is not a supported distro at this time".format(util.distro))
        return 1, 0
    util.WriteConsole(syslog.LOG_INFO, "Initiating Antivirus scan on {0}".format(target))
    err, msg = util.RunGetOutput(commands[util.distro])
    
    # exitCode == 0 -> Scan complete, no  errors or findings
    # exitCode == 1 -> Scan complete, found malware
    # exitCode >= 2 -> Error running scan
    if (err > 1):
        util.WriteLog(syslog.LOG_ERR, "Error [{0}] encountered while executing {1}".format(err, commands[util.distro]))
        return 1, 0
    # Parse output to gather:
    # - Names of malicious files (if any)
    # - Scanner engine version
    # - Elapsed time
    engineVer = "Unknown engine version"
    elapsedTime = "Unknown time"
    maliciousFiles = []
    lines = string.split(msg, "\n")
    for line in lines:
        if "FOUND" in line:
            lineSplit = string.split(line, ":")
            try:
                maliciousFiles.append(lineSplit[0])
            except ValueError:
                util.WriteLog(syslog.LOG_ERR, "Error parsing malware filename from output: {0}".format(line))
        elif "Engine version" in line:
            lineSplit = string.split(line, " ")
            try:
                engineVer = lineSplit[-1]
            except ValueError:
                util.WriteLog(syslog.LOG_ERR, "Error parsing engine version from output: {0}".format(line))
        elif "Time:" in line:
            lineSplit = string.split(line, " ")
            try:
                elapsedTime = "{0} {1}".format(lineSplit[1], lineSplit[2])
            except ValueError:
                util.WriteLog(syslog.LOG_ERR, "Error parsing elapsed time from output: {0}".format(line))
    # Set the log priority and message as appropriate
    logpri = syslog.LOG_WARNING if maliciousFiles else syslog.LOG_INFO
    resultMsg = "Antivirus scan complete, found {0} malicious files, {1}, {2}, exitcode={3}".format(len(maliciousFiles), engineVer, elapsedTime, err)
    if (len(maliciousFiles) > 0):
        resultMsg += ", {0}".format(";".join(maliciousFiles))
    util.WriteConsole(logpri, resultMsg)
    return 0, len(maliciousFiles)
    
# Performs AV scan by invoking shell script
# Returns a tuple containing
# 0 or 1 (if scan is successful/failed)
# Count of malicious files found
def RunScan(logFacility, args):
    syslog.openlog(facility=logFacility)
    target = args[argRootDir] if args.has_key(argRootDir) else defaultScanDir
    configFile = args[argConfigFile] if args.has_key(argConfigFile) else defaultConfigFile
    # Attempt to update signatures. Write warning if signatures are stale
    sigErr = UpdateSignatures(configFile)
    scanErr, findings = RunAVScan(target)
    syslog.closelog()
    return sigErr + scanErr, findings

